/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
/// <reference types="@webgpu/types" />
import './flags_webgpu';
import { backend_util, DataType, KernelBackend, Rank, RecursiveArray, ShapeMap, Tensor, Tensor1D, Tensor2D, Tensor3D, Tensor4D, TensorInfo, TimingInfo } from '@tensorflow/tfjs-core';
import { Glslang } from '@webgpu/glslang/dist/web-devel/glslang.onefile';
import { BufferManager } from './buffer_manager';
import * as webgpu_program from './kernels/webgpu_program';
export interface WebGPUMemoryInfo extends backend_util.MemoryInfo {
    numBytesInGPU: number;
    numBytesAllocatedInGPU: number;
    unreliable: boolean;
}
interface DataId {
}
export interface CPUTimerQuery {
    startMs: number;
    endMs: number;
}
export declare type WebGPUKernelInfo = {
    name: string;
    query: Promise<number>;
};
export declare type TimerNode = RecursiveArray<WebGPUKernelInfo> | WebGPUKernelInfo;
export interface WebGPUTimingInfo extends TimingInfo {
    uploadWaitMs: number;
    downloadWaitMs: number;
}
export declare class WebGPUBackend extends KernelBackend {
    device: GPUDevice;
    queue: GPUQueue;
    glslang: Glslang;
    commandQueue: GPUCommandEncoder[];
    private commandQueueOwnedIds;
    private binaryCache;
    private fromPixels2DContext;
    private bufferManager;
    private tensorMap;
    private tensorDisposalQueue;
    private uniformDisposalQueue;
    private disposed;
    private programTimersStack;
    private activeTimers;
    private uploadWaitMs;
    private downloadWaitMs;
    private cpuBackend;
    constructor(device: GPUDevice, glslang: Glslang);
    floatPrecision(): 32;
    flushDisposalQueue(): void;
    disposeData(dataId: DataId): void;
    memory(): WebGPUMemoryInfo;
    getBufferManager(): BufferManager;
    private acquireBuffer;
    private maybeReleaseBuffer;
    write(values: backend_util.BackendValues, shape: number[], dtype: DataType): DataId;
    move(dataId: DataId, values: backend_util.BackendValues, shape: number[], dtype: DataType): void;
    private submitQueue;
    getBuffer(dataId: DataId): GPUBuffer;
    private getBufferData;
    private convertAndCacheOnCPU;
    readSync(dataId: object): backend_util.BackendValues;
    read(dataId: object): Promise<backend_util.BackendValues>;
    time(f: () => void): Promise<WebGPUTimingInfo>;
    private getAndSavePipeline;
    private makeOutputArray;
    private tensorToBinding;
    startTimer(): {
        startMs: number;
        endMs: number;
    };
    endTimer(query: CPUTimerQuery): CPUTimerQuery;
    getQueryTime(query: CPUTimerQuery): Promise<number>;
    private uploadToGPU;
    compileAndRun<K extends TensorInfo>(program: webgpu_program.WebGPUProgram, inputs: TensorInfo[], output?: TensorInfo, programUniforms?: number[]): K;
    private makeUniforms;
    private getCPUBackend;
    private shouldExecuteOnCPU;
    pad<T extends Tensor>(x: T, paddings: Array<[number, number]>, constantValue: number): T;
    avgPool(x: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    maxPool(x: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    private binaryOp;
    add(a: Tensor, b: Tensor): Tensor;
    subtract(a: Tensor, b: Tensor): Tensor;
    private binaryCompareOp;
    less(a: Tensor, b: Tensor): Tensor;
    lessEqual(a: Tensor, b: Tensor): Tensor;
    greater(a: Tensor, b: Tensor): Tensor;
    greaterEqual(a: Tensor, b: Tensor): Tensor;
    private conv2dWithIm2Col;
    private conv2dByMatMul;
    conv2d(x: Tensor4D, filter: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    depthwiseConv2D(x: Tensor4D, filter: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    mapActivationToShaderProgram(activation: backend_util.Activation, packed?: boolean): string;
    fusedConv2d({ input, filter, convInfo, bias, activation, preluActivationWeights }: backend_util.FusedConv2DConfig): Tensor4D;
    private argMinMaxReduce;
    argMin(x: Tensor, axis: number): Tensor;
    argMax(x: Tensor, axis: number): Tensor;
    private reduce;
    max(x: Tensor, axes: number[]): Tensor;
    min(x: Tensor, axes: number[]): Tensor;
    sum(x: Tensor, axes: number[]): Tensor;
    clip<T extends Tensor>(x: T, min: number, max: number): T;
    slice<T extends Tensor>(x: T, begin: number[], size: number[]): T;
    stridedSlice<T extends Tensor>(x: T, begin: number[], end: number[], strides: number[]): T;
    concat(tensors: Tensor[], axis: number): Tensor;
    multiply(a: Tensor, b: Tensor): Tensor;
    floorDiv(a: Tensor, b: Tensor): Tensor;
    maximum(a: Tensor, b: Tensor): Tensor;
    neg<T extends Tensor>(x: T): T;
    tanh<T extends Tensor>(x: T): T;
    exp<T extends Tensor>(x: T): T;
    softmax<T extends Tensor>(logits: T, dim: number): T;
    log<T extends Tensor>(x: T): T;
    sigmoid<T extends Tensor>(x: T): T;
    relu<T extends Tensor>(x: T): T;
    relu6<T extends Tensor>(x: T): T;
    abs<T extends Tensor>(x: T): T;
    prelu<T extends Tensor>(x: T, alpha: T): T;
    select(condition: Tensor, a: Tensor, b: Tensor): Tensor;
    cropAndResize(image: Tensor4D, boxes: Tensor2D, boxIndex: Tensor1D, cropSize: [number, number], method: 'bilinear' | 'nearest', extrapolationValue: number): Tensor4D;
    fill<R extends Rank>(shape: ShapeMap[R], value: number | string, dtype?: DataType): Tensor<R>;
    zerosLike<R extends Rank>(x: Tensor<R>): Tensor<R>;
    resizeBilinear(x: Tensor4D, newHeight: number, newWidth: number, alignCorners: boolean): Tensor4D;
    reshape<R extends Rank>(x: Tensor, shape: ShapeMap[R]): Tensor<R>;
    cast<T extends Tensor>(x: T, dtype: DataType): T;
    transpose<T extends Tensor>(x: T, perm: number[]): T;
    batchToSpaceND<T extends Tensor>(x: T, blockShape: number[], crops: number[][]): T;
    spaceToBatchND<T extends Tensor>(x: T, blockShape: number[], paddings: Array<[number, number]>): T;
    batchMatMul(a: Tensor3D, b: Tensor3D, transposeA: boolean, transposeB: boolean): Tensor3D;
    fromPixels(pixels: backend_util.PixelData | ImageData | HTMLImageElement | HTMLCanvasElement | HTMLVideoElement, numChannels: number): Tensor3D;
    numDataIds(): number;
    dispose(): void;
}
export {};
