/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
import { util } from '@tensorflow/tfjs-core';
import { computeDispatch, flatDispatchLayout } from '../webgpu_util';
export class Im2ColProgram {
    constructor(outputShape, inputShape, convInfo) {
        this.variableNames = ['A'];
        this.workPerThread = 4;
        this.workGroupSize = [64, 1, 1];
        this.needsShapesUniforms = true;
        this.outputShape = outputShape;
        this.rank = outputShape.length;
        const size = util.sizeFromShape(this.outputShape);
        this.dispatchLayout = flatDispatchLayout(this.outputShape);
        this.dispatch = computeDispatch(this.dispatchLayout, this.outputShape, this.workGroupSize, [this.workPerThread, 1, 1]);
        const { filterWidth, inChannels, strideWidth, strideHeight, padInfo, outWidth, dilationWidth, dilationHeight, dataFormat } = convInfo;
        const { left, top } = padInfo;
        const itemsPerBlockRow = inChannels * filterWidth;
        const isChannelsLast = dataFormat === 'channelsLast';
        const rowDim = isChannelsLast ? 0 : 1;
        const colDim = isChannelsLast ? 1 : 2;
        this.userCode = `
      void main() {
        int index = int(gl_GlobalInvocationID.x);

        for(int i=0; i<${this.workPerThread}; i++) {
          int flatIndex = index * ${this.workPerThread} + i;

          ivec2 rc = getCoordsFromFlatIndex(flatIndex);

          if(flatIndex < ${size}) {
            int blockIndex = rc[0];
            int pos = rc[1];

            int offsetY = int(blockIndex / ${outWidth}) * ${strideHeight} -
              ${top};
            int d0 = offsetY + ${dilationHeight} * (pos / ${itemsPerBlockRow});
            float value = 0.0;
            if(d0 < ${inputShape[rowDim]} && d0 >= 0) {
              int offsetX = int(mod(float(blockIndex), ${outWidth}.) *
                ${strideWidth}. - ${left}.);
              int d1 = offsetX + ${dilationWidth} * (int(mod(float(pos),
                ${itemsPerBlockRow}.) / ${inChannels}.));
              int ch = int(mod(float(pos), ${inChannels}.));
              if(d1 < ${inputShape[colDim]} && d1 >= 0) {
                value = getA(d0, d1, ch);
              }
            }
            setOutput(flatIndex, value);
          }
        }
      }
    `;
    }
}
//# sourceMappingURL=im2col_webgpu.js.map