/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
import * as shader_preprocessor from '../shader_preprocessor';
export const makeBindGroup = (device, bindGroupLayout, inputs, output, uniforms) => {
    const bindings = [output, ...inputs];
    if (uniforms) {
        bindings.push(uniforms);
    }
    return device.createBindGroup({
        layout: bindGroupLayout,
        entries: bindings.map((b, i) => ({ binding: i, resource: b.resource })),
    });
};
const makeBindGroupLayout = (device, inputs, output, uniforms) => {
    const bindings = Array(1 + inputs.length)
        .fill({
        visibility: GPUShaderStage.COMPUTE,
        type: 'readonly-storage-buffer'
    }, 1);
    bindings[0] = {
        visibility: GPUShaderStage.COMPUTE,
        type: 'storage-buffer'
    };
    if (uniforms) {
        bindings.push({
            visibility: GPUShaderStage.COMPUTE,
            type: 'uniform-buffer'
        });
    }
    return device.createBindGroupLayout({
        entries: bindings.map((b, i) => (Object.assign({ binding: i }, b))),
    });
};
export const compileProgram = (glslang, device, program, inputsData, output, uniforms) => {
    const outputData = { dtype: output.dtype, shape: output.shape };
    const source = shader_preprocessor.makeShader(inputsData, outputData, program);
    const result = glslang.compileGLSLZeroCopy(source, 'compute', false);
    if (result.data.length === 0) {
        throw new Error('Shader compilation failed');
    }
    const bindGroupLayout = makeBindGroupLayout(device, inputsData, output, uniforms);
    const layout = device.createPipelineLayout({ bindGroupLayouts: [bindGroupLayout] });
    const module = device.createShaderModule({ code: result.data });
    const pipeline = device.createComputePipeline({ layout, computeStage: { module, entryPoint: 'main' } });
    result.free();
    return { bindGroupLayout, pipeline };
};
export function makeShaderKey(program, shapes, types) {
    const key = (program.workGroupSize ? program.workGroupSize.join(',') : '') +
        shapes.join(',') + types.join(',') + program.variableNames.join(',') +
        (program.shaderKey ? program.shaderKey : program.userCode);
    return key;
}
//# sourceMappingURL=webgpu_program.js.map